SystemOrganization addCategory: #'LRSTM-Core'!
SystemOrganization addCategory: 'LRSTM-Core-Compiler'!

OCASTSemanticAnalyzer subclass: #LRSTMTransformer
	instanceVariableNames: 'doFullTransformation doLocalTransformation'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LRSTM-Core-Compiler'!

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 2/3/2014 16:36'!
applySTMTransformations: methodNode with: aCompilationContext

	"We have two cases:
	  Either it is a normal method, and we do only the atomic blocks,
	  or it is a fully atomic method and we transform everything.
	
		doFullTransformation indicates that we do not only local adaptations.
	"
	(self pragmasForbidTransformation: methodNode)
		ifTrue: [^self].
	
	self compilationContext: aCompilationContext.
	self analyze: methodNode! !

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 2/3/2014 15:56'!
createAtomicAssignment: anAssignmentNode
	| var |
	var := anAssignmentNode variable.
	var isInstanceBinding ifTrue: [
		| selfRef instIdx sendArgs |
		selfRef := RBSelfNode named: 'self'.
		selfRef binding: (var binding scope lookupVar: 'self').
		
		instIdx := RBLiteralValueNode value: (var binding index).
		
		sendArgs := OrderedCollection new: 2.
		sendArgs add: instIdx.
		sendArgs add: anAssignmentNode value.
		^ RBMessageNode receiver: selfRef selector: #atomicInstVarAt:put: arguments: sendArgs.
	]
	ifFalse: [
		| globalAssoc sendArgs |
		var isLiteralBinding ifFalse: [self halt: 'I guess we still need to implement something, e.g., remote vars.'].
		
		globalAssoc := RBLiteralValueNode value: var binding assoc.
		
		sendArgs := OrderedCollection new: 1.
		sendArgs add: anAssignmentNode value.
		^ RBMessageNode receiver: globalAssoc selector: #atomicValue: arguments: sendArgs.
	].! !

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 2/3/2014 15:55'!
createAtomicReadNode: aVariableNode
	| selfRef instIdx sendArgs |
	
	aVariableNode isInstanceBinding ifTrue: [
		selfRef := RBSelfNode named: 'self'.
		selfRef binding: (aVariableNode binding scope lookupVar: 'self').
		
		instIdx := RBLiteralValueNode value: (aVariableNode binding index).
		sendArgs:= OrderedCollection new: 1.
		sendArgs add: instIdx.
		^ RBMessageNode receiver: selfRef selector: #atomicInstVarAt: arguments: sendArgs.	
	]
	ifFalse: [
		| globalAssoc |
		aVariableNode isLiteralBinding ifFalse: [self halt: 'I guess we still need to implement something, e.g., remote vars.'].
		
		globalAssoc := RBLiteralValueNode value: aVariableNode binding assoc.
		^ RBMessageNode receiver: globalAssoc selector: #atomicValue.
	].
	! !

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/3/2011 17:08'!
doTransformation
	^ doFullTransformation or: [doLocalTransformation]! !

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/3/2011 17:56'!
fullTransformation: aBool
	doFullTransformation := aBool! !

!LRSTMTransformer methodsFor: 'initialization' stamp: 'StefanMarr 12/4/2011 14:52'!
initialize
	super initialize.
	doFullTransformation  := false.
	doLocalTransformation := false.
! !

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/13/2013 20:32'!
pragmasForbidTransformation: methodNode
	methodNode pragmas do: [:each |
		((each selector = #atomicDoNotTransform   or: [
		  each selector = #atomic:])              or: [
		  each selector = #omniUnenforced])
			ifTrue: [^true].
	].
	^false.! !

!LRSTMTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 2/3/2014 12:38'!
transformMessageNode: aMessageNode

	| selector |
	selector := aMessageNode selector.
	(selector = #atomic or: [aMessageNode isInlined]) ifFalse: [
		aMessageNode selector: selector asAtomicSelector.
	].
	
	super visitMessageNode: aMessageNode.! !

!LRSTMTransformer methodsFor: 'visitor' stamp: 'StefanMarr 2/3/2014 11:30'!
visitAssignmentNode: anAssignmentNode
	super visitAssignmentNode: anAssignmentNode.
	
	self doTransformation ifFalse: [^self].
	
	(anAssignmentNode variable isInstanceBinding or:
	[anAssignmentNode variable isLiteralBinding "or: REMOTE NOT YET SUPPORTED
	[anAssignmentNode variable isRemote]"]) ifTrue: [
		| replacement |
		replacement := self createAtomicAssignment: anAssignmentNode.
		anAssignmentNode parent replaceNode: anAssignmentNode withNode: replacement.
	]! !

!LRSTMTransformer methodsFor: 'visitor' stamp: 'StefanMarr 2/3/2014 11:31'!
visitMessageNode: aMessageNode
	| doLocalTransformationOld |
	doLocalTransformationOld := doLocalTransformation.
	
	doLocalTransformation := doLocalTransformation or: 
								 [aMessageNode selector = #atomic or:
								 [aMessageNode selector = #atomicIfConflict: or:
								 [aMessageNode selector = #atomicRetry           ]]].

	(self doTransformation and: [aMessageNode isInlined not])
		ifTrue:  [self  transformMessageNode: aMessageNode.]
		ifFalse: [super visitMessageNode:    aMessageNode.].
	
	doLocalTransformation := doLocalTransformationOld.! !

!LRSTMTransformer methodsFor: 'visitor' stamp: 'StefanMarr 2/3/2014 11:31'!
visitMethodNode: aMethodNode
	super visitMethodNode: aMethodNode.
	doFullTransformation ifTrue: [
		aMethodNode selector: (aMethodNode selector asAtomicSelector)
	].! !

!LRSTMTransformer methodsFor: 'visitor' stamp: 'StefanMarr 2/3/2014 15:53'!
visitVariableNode: aVariableNode
	| replacement |
	super visitVariableNode: aVariableNode.
	
	self doTransformation ifFalse: [^self].
	
	
	"Not every variable node we are going to encounter is a read.
	 Make sure it is not the variable node of a parent assignment node.
	 Furthermore, we do not need to take care of temp assignments.
	 Only globals, instances, and perhaps remote onces of a closure."
	(aVariableNode isRead
	 and: [aVariableNode isLiteralBinding   or: 
		   [aVariableNode isInstanceBinding "or: REMOTES NOT SUPPORTED at the moment
			[aVariableNode isRemote]"]])
		ifFalse: [^self].
	
	"Now we are sure that we have to transform a variable read".
	replacement := self createAtomicReadNode: aVariableNode.
	aVariableNode parent replaceNode: aVariableNode withNode: replacement. 
! !

!LookupKey methodsFor: '*LRSTM-Core-atomic' stamp: 'lr 5/25/2007 11:18'!
atomicValue
	^ self workingCopy value! !

!LookupKey methodsFor: '*LRSTM-Core-atomic' stamp: 'lr 5/25/2007 11:18'!
atomicValue: anObject
	^ self workingCopy value: anObject! !

!CompiledMethod methodsFor: '*LRSTM-Core-Compiler' stamp: 'lr 5/9/2007 10:37'!
atomicMethod
	^ LRSTMCompiler atomicMethodFor: self! !

OpalCompiler subclass: #LRSTMCompiler
	instanceVariableNames: 'doFullTransformation'
	classVariableNames: 'ErrorOnMissingAtomicMethod OldCompiler'
	poolDictionaries: ''
	category: 'LRSTM-Core-Compiler'!

!LRSTMCompiler commentStamp: 'StefanMarr 11/28/2011 22:47' prior: 0!
This is just a helper class holding the compilation related methods similar to the original implementation of ACCompiler!

!LRSTMCompiler class methodsFor: 'actions' stamp: 'StefanMarr 6/4/2013 13:43'!
atomicFail: aCompiledMethod
	"Failed to compile aCompiledMethod. Show error in transcript and install the origianl method."
	<omniUnenforced>
	
	Transcript
		show: 'ERROR: Failed to compile for atomic execution: '; show: aCompiledMethod methodClass; 
		show: '>>#'; show: aCompiledMethod selector; cr.
	^ aCompiledMethod! !

!LRSTMCompiler class methodsFor: 'actions' stamp: 'StefanMarr 6/4/2013 13:43'!
atomicMethod: aMessage missing: aReceiver
	"This method is called whenever a missing atomic method is called. Installs a whole hierarchy of atomic methods."
	<omniUnenforced>
	^ self atomicMethod: aMessage missing: aReceiver in: aReceiver class ! !

!LRSTMCompiler class methodsFor: 'actions' stamp: 'StefanMarr 7/15/2013 17:11'!
atomicMethod: aMessage missing: aReceiver in: aClass
	"This method is called whenever a missing atomic method is called. Installs a whole hierarchy of atomic methods."
	<omniUnenforced>
	
	| selector addedAtomicMethod curClass |
	selector := aMessage selector asNormalSelector.
	
	ErrorOnMissingAtomicMethod ifTrue: [
		^ (MessageNotUnderstood new
				message: aMessage;
				receiver: self;
				signal)].

	(ProtoObject withAllSubclasses select: [:c | c includesSelector: selector])
	  do: [ :class |
			self installAtomicVersionOf: selector in: class ]
	  displayingProgress: 'Atomicify Methods: ', selector.

! !

!LRSTMCompiler class methodsFor: 'helper' stamp: 'StefanMarr 2/3/2014 15:35'!
devHelperButtons
	| b |
	b:= ButtonModel new.
	b label: 'Restore OPAL'.
	b action: [ SmalltalkImage compilerClass: OpalCompiler. ].
	b openWithSpec.! !

!LRSTMCompiler class methodsFor: 'compiler' stamp: 'StefanMarr 2/3/2014 15:24'!
disableCompiler
	SmalltalkImage compilerClass: OldCompiler.! !

!LRSTMCompiler class methodsFor: 'actions' stamp: 'StefanMarr 9/12/2012 13:55'!
disableErrorOnMissingAtomicMethod
	ErrorOnMissingAtomicMethod := false.! !

!LRSTMCompiler class methodsFor: 'compiler' stamp: 'StefanMarr 2/3/2014 15:24'!
enableCompiler
	OldCompiler := Smalltalk compilerClass.
	OldCompiler ifNil: [ OldCompiler := OpalCompiler ].
	
	SmalltalkImage compilerClass: self.! !

!LRSTMCompiler class methodsFor: 'actions' stamp: 'StefanMarr 9/12/2012 13:55'!
errorOnMissingAtomicMethod
	ErrorOnMissingAtomicMethod := true.! !

!LRSTMCompiler class methodsFor: 'compiler' stamp: 'StefanMarr 12/3/2011 22:05'!
evaluate: textOrString notifying: aController logged: logFlag 
	"This is used for DoIts, the do not seem to be supported currently by OPAL. Let's use a plain compiler instead."
	Transcript cr; show: 'Did not use the LRSTM compiler to compile the following code:'.
	Transcript cr; show: '---'.
	Transcript cr; show: textOrString contents.
	Transcript cr; show: '---'.
	
	^ Compiler evaluate: textOrString notifying: aController logged: logFlag.! !

!LRSTMCompiler class methodsFor: 'compiler' stamp: 'StefanMarr 6/4/2013 14:00'!
flushAllAtomicMethods
	Smalltalk allClassesAndTraits 
		do: [ :ea | ea flushAtomic. ea class flushAtomic ]
		displayingProgress: 'Flushing Atomic Methods'.! !

!LRSTMCompiler class methodsFor: 'class initialization' stamp: 'StefanMarr 6/11/2013 19:41'!
initializeLRSTM
	Transcript cr; show: '===INITIALIZE LRSTM [START]===='.
	
	ErrorOnMissingAtomicMethod := false.
	
	self recompileClassesWithLRSTMCompiler.
	
	SystemChangeNotifier uniqueInstance
		notify: self ofSystemChangesOfItem: #method change: #Added using: #onMethodAdded:;
		notify: self ofSystemChangesOfItem: #method change: #Modified using: #onMethodAdded:;
		notify: self ofSystemChangesOfItem: #method change: #Removed using: #onMethodRemoved:.
	
	Transcript cr; show: '===INITIALIZE LRSTM [COMPLETED]===='.
! !

!LRSTMCompiler class methodsFor: 'accessing' stamp: 'StefanMarr 7/15/2013 17:16'!
installAtomicVersionOf: aSelector in: aClass
	"Answer the atomic counterpart of aCompiledMethod.
	Methods with the pragma #atomicDoNotTransform are considered to be already atomic.
	Methods with the pragma #atomic: will use the annotation code for the atomic counterpart.
	#atomicInstead: will use the annotiation code for atomic transformation."
	<omniUnenforced>

	| atomicCM atomicName newMethod compiledMethod |
	compiledMethod := aClass lookupSelector: aSelector.
	
	"Is already atomic do nothing"
	aSelector isAtomicSelector ifTrue: [
		^ compiledMethod ].

	atomicName := aSelector asAtomicSelector.
	
	compiledMethod pragmas do: [ :each |
		each keyword = #omniUnenforced
			ifTrue: [ atomicCM := compiledMethod.
				self installOrExchange: atomicCM in: aClass for: atomicName.
				^ atomicCM ].
			
		each keyword = #omniReplacement:
			ifTrue: [
				| replacementName replacementMeth |
				replacementName := each arguments first.
				(replacementName beginsWith: #omni)
					ifTrue: ["that should be omniBaseNew etc, then just normally process method, i.e., fall through to end of method. Those methods are specific to our Omni implementation, and not relevant for LRSTM."]
					ifFalse:[
							(replacementName beginsWith: #atomic) ifFalse: [
								replacementName := (#atomic, replacementName) asSymbol].
				
							replacementMeth := aClass lookupSelector: replacementName.
							atomicCM := self installAtomicVersionOf: replacementName in: replacementMeth methodClass.
							
							self installOrExchange: atomicCM in: aClass for: atomicName.
							^ atomicCM ]]].
	
	newMethod := LRSTMCompiler recompileAtomically: compiledMethod.
	self installOrExchange: newMethod in: aClass for: atomicName.
							
	^ newMethod! !

!LRSTMCompiler class methodsFor: 'helper' stamp: 'StefanMarr 7/15/2013 16:57'!
installOrExchange: compiledMethod in: class for: aSelector
	(class includesLocalSelector: aSelector)
		ifFalse: [ self registerLocalSelector: aSelector ].
	
	class methodDict at: aSelector put: compiledMethod.
	
	aSelector flushCache.
	compiledMethod flushCache.
	^ compiledMethod! !

!LRSTMCompiler class methodsFor: 'events' stamp: 'StefanMarr 6/4/2013 13:46'!
onMethodAdded: anObject
	<omniUnenforced>
	
	"Ignore the already atomic methods"
	anObject itemSelector isAtomicSelector ifTrue: [
		^self].
	
	"Also blacklist a few things"
	((anObject itemClass = Debugger)
	or: [anObject itemClass = Decompiler])
		ifTrue: [^ self].

	"For the rest we do the atomic stuff"	
	self installAtomicVersionOf: (anObject itemClass methodDict
			at: anObject itemSelector).! !

!LRSTMCompiler class methodsFor: 'events' stamp: 'StefanMarr 6/4/2013 13:46'!
onMethodRemoved: anObject
	<omniUnenforced>
	
	anObject itemClass methodDict 
		removeKey: anObject itemSelector 
			asAtomicSelector
		ifAbsent: nil! !

!LRSTMCompiler class methodsFor: 'compiler' stamp: 'StefanMarr 6/4/2013 13:46'!
recompile: aCompiledMethod
	"Recompile the given method normally"
	<omniUnenforced>
	
	| compiler |
	compiler := LRSTMCompiler new.
	^ compiler recompile: aCompiledMethod.! !

!LRSTMCompiler class methodsFor: 'compiler' stamp: 'StefanMarr 6/4/2013 13:46'!
recompileAtomically: aCompiledMethod
	"Recompile the given method normally"
	<omniUnenforced>
	
	| compiler |
	compiler := LRSTMCompiler new.
	^ compiler recompileAtomically: aCompiledMethod.! !

!LRSTMCompiler class methodsFor: 'compiler' stamp: 'StefanMarr 6/6/2013 16:43'!
recompileClassesWithLRSTMCompiler
	<omniUnenforced>
	
	self enableCompiler.
	
	(#(O LRSTM) gather: [:clsName |
	(PackageInfo named: clsName) classes ])
		do: [:c | c compileAll]
		displayingProgress: 'Recompiling with LRSTM support enabled'.! !

!LRSTMCompiler class methodsFor: 'class initialization' stamp: 'StefanMarr 6/6/2013 16:37'!
unload
	self flushAllAtomicMethods.! !

!LRSTMCompiler class methodsFor: 'class initialization' stamp: 'StefanMarr 12/2/2011 11:42'!
unregisterChangeListener

	SystemChangeNotifier uniqueInstance
		noMoreNotificationsFor: self.! !

!LRSTMCompiler methodsFor: 'private' stamp: 'StefanMarr 2/3/2014 12:38'!
doSemanticAnalysis
	| transformer |

	super doSemanticAnalysis.
	
	transformer := LRSTMTransformer new.
		transformer fullTransformation: doFullTransformation.
		transformer applySTMTransformations: ast with: compilationContext.! !

!LRSTMCompiler methodsFor: 'initialize-release' stamp: 'StefanMarr 12/3/2011 18:02'!
initialize
	super initialize.
	doFullTransformation := false.! !

!LRSTMCompiler methodsFor: 'private' stamp: 'StefanMarr 12/4/2011 15:44'!
parse: methodBoolean
	| methodNode transformer |
	methodNode := super parse: methodBoolean.
	
	methodBoolean ifTrue: [
		transformer := LRSTMTransformer new.
		transformer fullTransformation: doFullTransformation.
		transformer applySTMTransformations: methodNode inScopeOf: class.
	].
	^ methodNode! !

!LRSTMCompiler methodsFor: 'atomicity support' stamp: 'StefanMarr 12/3/2011 18:06'!
recompile: aCompiledMethod
	"Recompile the given method normally, i.e., with out full STM transformation"

	^ self recompile: aCompiledMethod atomically: false.! !

!LRSTMCompiler methodsFor: 'atomicity support' stamp: 'StefanMarr 2/3/2014 16:35'!
recompile: aCompiledMethod atomically: aBool
	"Recompile the given method with the STM enabled"

	| methodSource newSelector  newMethod |

	doFullTransformation := aBool.
	
	methodSource := aCompiledMethod getSource.
	newSelector := aCompiledMethod selector asAtomicSelector.
		
	newMethod := self
		compile: methodSource in: aCompiledMethod methodClass notifying: nil
		ifFail: [^ self problem: 'syntax error' sel: newSelector in: aCompiledMethod methodClass].

	^ newMethod generate.! !

!LRSTMCompiler methodsFor: 'atomicity support' stamp: 'StefanMarr 12/3/2011 18:06'!
recompileAtomically: aCompiledMethod
	"Recompile the given method with the STM enabled"

	^ self recompile: aCompiledMethod atomically: true.! !

!RBMessageNode class methodsFor: '*LRSTM-Core-Compiler' stamp: 'lr 5/10/2007 11:43'!
receiver: aValueNode atomicSelector: aSymbol arguments: anotherValueNodes
	^ self new
		receiver: aValueNode;
		arguments: anotherValueNodes;
		atomicSelector: aSymbol;
		yourself! !

!RBMessageNode methodsFor: '*LRSTM-Core-Compiler' stamp: 'lr 5/9/2007 14:50'!
atomicSelector: aSelector
	| keywords numArgs |
	aSelector isAtomicSelector 
		ifTrue: [ self error: 'Normal selector expected: ' , aSelector ].
	selector := aSelector asAtomicSelector.
	keywords := aSelector keywords.
	numArgs := selector last = $:
		ifTrue: [ keywords size ]
		ifFalse: [ 0 ].
	selectorParts := numArgs = 0
		ifTrue: [ Array with: (RBIdentifierToken value: selector start: nil) ]
		ifFalse: [ keywords collect: [ :each | RBKeywordToken value: each start: nil ] ]! !

!RBMessageNode methodsFor: '*LRSTM-Core-Compiler-inline' stamp: 'JorgeRessia 11/25/2009 13:33'!
isInlineAndOr
	self receiver isBlock ifTrue: [^ false].
	self isCascaded ifTrue: [^ false].
	(self selectorParts allSatisfy: [ :each | #(and: or:) includes: each value ]) 
		ifFalse: [^ false].
	(self arguments allSatisfy: [ :each | each isBlock ])  ifFalse: [^ false].
	(self arguments allSatisfy: [ :each | each arguments isEmpty ])
			ifFalse: [
				self notify: 'and: (or:) takes zero-arg block'.
				^ false ].
	^ true! !

!RBMessageNode methodsFor: '*LRSTM-Core-Compiler-inline' stamp: 'jorgeRessia 11/20/2009 16:40'!
isInlineCase

	self isCascaded ifTrue: [^ false].
	(#(caseOf: caseOf:otherwise:) includes: self selector) ifFalse: [^ false].
	self arguments size = 2 ifTrue: [  "otherwise block"
		self arguments last isBlock ifFalse: [^ false]].
	self arguments first isArray ifFalse: [^ false].
	self arguments first statements do: [:assoc |
		(assoc isMessage and: [assoc selector == #->]) ifFalse: [^ false].
		assoc receiver isBlock ifFalse: [^ false].
		assoc receiver arguments isEmpty ifFalse: [self notify: 'caseOf: takes zero-arg blocks'. ^ false].
		assoc arguments first isBlock ifFalse: [^ false].
		assoc arguments first arguments isEmpty ifFalse: [self notify: 'caseOf: takes zero-arg blocks'].
	].
	^ true! !

!RBMessageNode methodsFor: '*LRSTM-Core-Compiler-inline' stamp: 'jorgeRessia 11/20/2009 16:40'!
isInlineIf

	self receiver isBlock ifTrue: [^ false].
	self isCascaded ifTrue: [^ false].
	(#(ifTrue: ifFalse: ifTrue:ifFalse: ifFalse:ifTrue:) includes: self selector) ifFalse: [^ false].
	self arguments do: [:node | node isBlock ifFalse: [^ false]].
	self arguments do: [:block |
		block arguments isEmpty ifFalse: [
			self notify: 'ifTrue:ifFalse: takes zero-arg blocks'.
			^ false
		]
	].
	^ true! !

!RBMessageNode methodsFor: '*LRSTM-Core-Compiler-inline' stamp: 'MarcusDenker 8/21/2010 13:40'!
isInlineIfNil

	| assertNone assertOneOrNone |
	self receiver isBlock ifTrue: [^ false].
	self isCascaded ifTrue: [^ false].
	(#(ifNil: ifNotNil: ifNil:ifNotNil: ifNotNil:ifNil:) includes: self selector) ifFalse: [^ false].
	self arguments do: [:node | node isBlock ifFalse: [^ false]].
	assertNone := [:block | 
		block arguments isEmpty ifFalse: [self notify: 'ifNil: takes zero-arg block'. ^ false]
	].
	assertOneOrNone := [:block | 
		block arguments size > 1 ifTrue: [self notify: 'ifNotNil: takes zero- or one-arg block'. ^ false]
	].
	self selector caseOf: {
		[#ifNil:] 		 	-> [assertNone value: self arguments first].
		[#ifNil:ifNotNil:] 	-> [assertNone value: self arguments first. 
							   assertOneOrNone value: self arguments last].
		[#ifNotNil:] 		  	-> [assertOneOrNone value: self arguments first].
		[#ifNotNilDo:] 		  	-> [assertOneOrNone value: self arguments first].
		[#ifNotNil:ifNil:] 	-> [assertOneOrNone value: self arguments first. 
							   assertNone value: self arguments last]
	}.
	^ true! !

!RBMessageNode methodsFor: '*LRSTM-Core-Compiler-inline' stamp: 'MarcusDenker 9/3/2010 15:03'!
isInlineToDo

	| block step |
	self receiver isBlock ifTrue: [^ false].
	self isCascaded ifTrue: [^ false].
	(#(to:do: to:by:do:) includes: self selector) ifFalse: [^ false].
	self arguments first isBlock ifTrue: [^ false].
	block := self arguments last.
	block isBlock ifFalse: [^ false].
	block arguments size = 1 ifFalse: [
		self notify: 'to:do: block must take one arg'. ^ false].
	(RBParseTreeSearcher new
		matches: block arguments first name , ' := `@object' do: [:n :a | true];
		executeTree: block initialAnswer: false) ifTrue: [^ false].
	self arguments size = 3 "to:by:do:" ifTrue: [
		step := self arguments second.
		step isLiteral ifFalse: [^ false].
		step value = 0 ifTrue: [self notify: 'by: step can''t be zero'. ^ false].
	].
	^ true! !

!RBMessageNode methodsFor: '*LRSTM-Core-Compiler-inline' stamp: 'jorgeRessia 11/20/2009 16:40'!
isInlineWhile

	self isCascaded ifTrue: [^ false].
	(#(whileFalse: whileTrue: whileFalse whileTrue) includes: self selector) ifFalse: [^ false].
	self receiver isBlock ifFalse: [^ false].
	self receiver arguments isEmpty ifFalse: [self notify: 'while receiver block must have no arguments'. ^ false].
	self arguments isEmpty ifFalse: [
		self arguments first isBlock ifFalse: [^ false].
		self arguments first arguments isEmpty ifFalse: [self notify: 'while takes a zero-arg block as its argument'. ^ false].
	].
	^ true! !

!RBMessageNode methodsFor: '*LRSTM-Core-Compiler-inline' stamp: 'StefanMarr 12/3/2011 14:22'!
isInlined
	^ self isInlineAndOr or:
	 [self isInlineCase  or:
	 [self isInlineIf    or:
	 [self isInlineIfNil or:
	 [self isInlineToDo  or:
	  self isInlineWhile    ]]]]! !

!String methodsFor: '*LRSTM-Core-Compiler-override' stamp: 'StefanMarr 11/28/2011 20:53'!
numArgs 
	"Answer either the number of arguments that the receiver would take if considered a selector.  Answer -1 if it couldn't be a selector. It is intended mostly for the assistance of spelling correction."

	| firstChar numColons start ix |
	self size = 0 ifTrue: [^ -1].
	self isAtomicSelector 
		ifTrue: [ ^ self asNormalSelector numArgs ].
	firstChar := self at: 1.
	(firstChar isLetter or: [ firstChar = $_]) ifTrue:
		["Fast reject if any chars are non-alphanumeric
		NOTE: fast only for Byte things - Broken for Wide"
		self class isBytes
			ifTrue: [(self findSubstring: '~' in: self startingAt: 1 matchTable: Tokenish) > 0 ifTrue: [^ -1]]
			ifFalse: [2 to: self size do: [:i | (self at: i) tokenish ifFalse: [^ -1]]].
		"Fast colon count"
		numColons := 0.  start := 1.
		[(ix := self indexOf: $: startingAt: start) > 0]
			whileTrue:
				[ix = start ifTrue: [^-1].
				numColons := numColons + 1.
				start := ix + 1].
		numColons = 0 ifTrue: [^ 0].
		self last = $:
			ifTrue: [^ numColons]
			ifFalse: [^ -1]].
	firstChar isSpecial ifTrue:
		[self size = 1 ifTrue: [^ 1].
		2 to: self size do: [:i | (self at: i) isSpecial ifFalse: [^ -1]].
		^ 1].
	^ -1.! !

!Symbol methodsFor: '*LRSTM-Core-Compiler-override' stamp: 'lr 5/10/2007 12:01'!
precedence
	"Answer the receiver's precedence, assuming it is a valid Smalltalk
	message selector or 0 otherwise.  The numbers are 1 for unary,
	2 for binary and 3 for keyword selectors."

	self isAtomicSelector 
		ifTrue: [ ^ self asNormalSelector precedence ].
	self size = 0 ifTrue: [^ 0].
	self first isLetter ifFalse: [^ 2].
	self last = $: ifTrue: [^ 3].
	^ 1! !

!Behavior methodsFor: '*LRSTM-Core-Compiler' stamp: 'lr 5/9/2007 13:19'!
atomicMethods
	^ Array streamContents: [ :stream |
		self methodDictionary keysAndValuesDo: [ :key :value |
			key isAtomicSelector 
				ifTrue: [ stream nextPut: value ] ] ]! !

!Behavior methodsFor: '*LRSTM-Core-Compiler' stamp: 'lr 5/11/2007 16:01'!
flushAtomic
	| copy |
	self methodDictionary keysDo: [ :each |
		each isAtomicSelector ifTrue: [
			copy := copy ifNil: [ self methodDictionary copy ].
			copy removeDangerouslyKey: each ifAbsent: [] ] ].
	copy ifNotNil: [ self methodDictionary become: copy ]! !

!Behavior methodsFor: '*LRSTM-Core-Compiler-override' stamp: 'lr 5/9/2007 11:05'!
selectors
	"Answer a Set of all the message selectors specified in the receiver's 
	method dictionary."

	^ self methodDict keys reject: [ :each | each isAtomicSelector ]! !

!Behavior methodsFor: '*LRSTM-Core-Compiler-override' stamp: 'lr 5/9/2007 11:05'!
selectorsAndMethodsDo: aBlock
	"Evaluate selectorBlock for all the message selectors in my method dictionary."

	^ self methodDict keysAndValuesDo: [ :selector :method |
		selector isAtomicSelector 
			ifFalse: [ aBlock value: selector value: method ] ]! !

!Behavior methodsFor: '*LRSTM-Core-Compiler-override' stamp: 'lr 5/9/2007 11:06'!
selectorsDo: selectorBlock
	"Evaluate selectorBlock for all the message selectors in my method dictionary."

	^ self methodDict keysDo: [ :selector |
		selector isAtomicSelector 
			ifFalse: [ selectorBlock value: selector ] ]! !

!RBVariableNode methodsFor: '*LRSTM-Core-Compiler-ast-semantic-accessing' stamp: 'StefanMarr 2/3/2014 12:52'!
isInstanceBinding
	^ self binding isInstance! !

!RBVariableNode methodsFor: '*LRSTM-Core-Compiler-ast-semantic-accessing' stamp: 'StefanMarr 2/3/2014 12:51'!
isLiteralBinding
	^ self binding isLiteralVariable! !

!RBVariableNode methodsFor: '*LRSTM-Core-Compiler-ast-semantic-accessing' stamp: 'StefanMarr 12/4/2011 15:49'!
variableBinding: aBinding
	^ self propertyAt: #variableBinding put: aBinding! !

!RBMethodNode methodsFor: '*LRSTM-Core-Compiler' stamp: 'StefanMarr 12/3/2011 22:26'!
sourceText
	^source! !

!LRSTMTransaction methodsFor: '*LRSTM-Core-Model' stamp: 'StefanMarr 6/4/2013 13:45'!
do: aBlock ifConflict: aConflictBlock
	"Evaluate aBLock within a new transaction, unless we are already in an existing transactional context. Evaluate aConflictBlock if the transaction conflicts with concurrent edits."

	<omniUnenforced>
	| result |
	parent := Processor activeProcess currentTransaction.
	Processor activeProcess currentTransaction: self.
	context := thisContext sender.
	self begin.
	[ result := aBlock 
		ensure: [ Processor activeProcess currentTransaction: parent ] ]
		ifCurtailed: [ self abort ].
	self commitIfConflict: aConflictBlock.
	^ result! !

!TraitBehavior methodsFor: '*LRSTM-Core-Compiler' stamp: 'lr 5/11/2007 15:58'!
flushAtomic
	| changed copy |
	changed := false.
	copy := self methodDictionary copy.
	self methodDictionary keysDo: [ :each |
		each isAtomicSelector ifTrue: [ 
			copy removeDangerouslyKey: each ifAbsent: [].
			changed := true ] ].
	changed ifTrue: [ self methodDictionary become: copy ]! !

!TraitBehavior methodsFor: '*LRSTM-Core-Compiler-override' stamp: 'lr 5/9/2007 11:05'!
selectorsAndMethodsDo: aBlock
	"Evaluate selectorBlock for all the message selectors in my method dictionary."

	^ self methodDict keysAndValuesDo: [ :selector :method |
		selector isAtomicSelector 
			ifFalse: [ aBlock value: selector value: method ] ]! !

!Object methodsFor: '*LRSTM-Core-atomic' stamp: 'StefanMarr 6/4/2013 13:43'!
atomicAt: anInteger
	<omniUnenforced>
	^ self workingCopy at: anInteger! !

!Object methodsFor: '*LRSTM-Core-atomic' stamp: 'StefanMarr 6/4/2013 13:43'!
atomicAt: anInteger put: anObject
	<omniUnenforced>
	^ self workingCopy at: anInteger put: anObject! !

!Object methodsFor: '*LRSTM-Core-atomic' stamp: 'StefanMarr 6/4/2013 13:43'!
atomicBasicAt: anInteger
	<omniUnenforced>
	^ self workingCopy basicAt: anInteger! !

!Object methodsFor: '*LRSTM-Core-atomic' stamp: 'StefanMarr 6/4/2013 13:43'!
atomicBasicAt: anInteger put: anObject
	<omniUnenforced>
	^ self workingCopy basicAt: anInteger put: anObject! !

!Object methodsFor: '*LRSTM-Core-atomic' stamp: 'StefanMarr 6/4/2013 13:43'!
atomicInstVarAt: anInteger
	<omniUnenforced>
	^ self workingCopy instVarAt: anInteger! !

!Object methodsFor: '*LRSTM-Core-atomic' stamp: 'StefanMarr 6/4/2013 13:43'!
atomicInstVarAt: anInteger put: anObject
	<omniUnenforced>
	^ self workingCopy instVarAt: anInteger put: anObject! !

!Object methodsFor: '*LRSTM-Core-atomic' stamp: 'StefanMarr 6/4/2013 13:44'!
atomicShallowCopy
	<omniUnenforced>
	^ self workingCopy shallowCopy! !

!Object methodsFor: '*LRSTM-Core-Compiler-overrides' stamp: 'StefanMarr 6/4/2013 13:45'!
doesNotUnderstand: aMessage 
	 "Handle the fact that there was an attempt to send the given message to the receiver but the receiver does not understand this message (typically sent from the machine when a message is sent to the receiver and no method is defined for that selector)."
	"Testing: (3 activeProcess)"
	"fixed suggested by Eliot miranda to make sure 
	
	[Object new blah + 1]
 		on: MessageNotUnderstood
 		do: [:e | e resume: 1] does not loop indefinitively"
		
	| exception resumeValue |
	
	<omniUnenforced>	
	aMessage selector isAtomicSelector
		ifTrue: [
			LRSTMCompiler atomicMethod: aMessage missing: self in: aMessage lookupClass.
			
			(self respondsTo: aMessage selector) ifTrue: [
				"And try again."
				^ aMessage sentTo: self.
			]].
		
	(exception := MessageNotUnderstood new)
				message: aMessage;
				receiver: self.
	resumeValue := exception signal.
	^exception reachedDefaultHandler
				ifTrue: [aMessage sentTo: self]
				ifFalse: [resumeValue]! !
